/* HuliFontViewer - A simple java font character map viewer.
 * Copyright (C) 2004-2005 Timothy Robinson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package org.huli.hulifontviewer;

import javax.swing.*;
import java.awt.*;
import java.awt.event.AdjustmentListener;
import java.awt.event.AdjustmentEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.ChangeEvent;
import java.awt.event.ItemListener;
import java.awt.event.ItemEvent;
import java.io.*;

public class DrawCharacterMap extends JComponent
    implements AdjustmentListener, ChangeListener, ItemListener {

    private int rows = 8;
    private int rowsPerPage = 8;
    private int cols = 1;
    private int boxWidth;
    private int boxHeight;
    private int horizontalBoxPadding = 10;
    private int verticalBoxPadding = 10;
    private int horizontalMarginPadding = 10;
    private int verticalMarginPadding = 10;
    private boolean showMetricLines = true;

    private int metricsLineExtra = 5;
    private int maxDescent;
    private int maxAscent;
    private int maxAdvance;

    private int offset = 0;

    private Font font = null;
    private float fontSize = 30.0f;
    private String fontFileName = null;

    private JScrollBar scrollBar = null;

    public void setScrollBar(JScrollBar s) {
        scrollBar = s;
    }

    public void setFontFileName(String s) {
        fontFileName = s;
        font = getNewFont();

        /*
        if (font != null) {
            rows = font.getNumGlyphs() / cols + 1;
            System.out.println("rows: " + font.getNumGlyphs() + " / " + cols + " + 1 = " + rows);
        } else {
            rows = rowsPerPage;
        }
        */

        scrollBar.setValue(0);
        //scrollBar.setMaximum(rows);
        scrollBar.setMaximum(500);
        repaint();
    }

    public void setFontSize(float f) {
        fontSize = f;
        font = getNewFont();
        repaint();
    }

    protected void paintComponent(Graphics g) {
        Color textColor = new Color(0, 0, 0);
        Color backgroundColor = new Color(255, 255, 255);
        Color baseColor = new Color(0, 0, 255);
        Color ascentColor = new Color(255, 0, 0);
        Color descentColor = new Color(0, 255, 0);

        // clear
        g.setColor(backgroundColor);
        g.fillRect(0, 0, getWidth(), getHeight());

        if (fontFileName == null) {
            return;
        }

        Graphics2D g2 = (Graphics2D) g;
        Font numFont = g2.getFont();

        FontMetrics metrics = g.getFontMetrics();
        int numFontMaxDescent = metrics.getMaxDescent();
        int numFontMaxAscent =  metrics.getMaxAscent();

        g2.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING ,
                            RenderingHints.VALUE_TEXT_ANTIALIAS_ON);

        if (font != null) {
            g2.setFont(font);

            metrics = g.getFontMetrics();
            /*
            System.out.println("in paint()");
            System.out.println("hasUniformLineMetrics(): " + metrics.hasUniformLineMetrics());
            System.out.println("maxDecent(): " + metrics.getMaxDescent());
            System.out.println("maxAscent(): " + metrics.getMaxAscent());
            System.out.println("maxAdvance(): " + metrics.getMaxAdvance());
            */
            /*
            */

            maxDescent = metrics.getMaxDescent();
            maxAscent =  metrics.getMaxAscent();
            maxAdvance =  metrics.getMaxAdvance();

            boxWidth = maxAdvance + (2 * horizontalBoxPadding);
            boxHeight = maxDescent + maxAscent + numFontMaxAscent +
                numFontMaxDescent + (2 * verticalBoxPadding); 
            /*
            System.out.println("boxWidth: " + boxWidth);
            System.out.println("boxHeight: " + boxHeight);
            */

            cols = 1;
            while (cols * boxWidth < getWidth()) {
                cols++;
            }
            cols--;

            rows = 1;
            while (rows * boxHeight < getHeight()) {
                rows++;
            }
            rows--;

            horizontalMarginPadding = (getWidth() - cols * boxWidth) / 2;
            verticalMarginPadding = (getHeight() - rows * boxHeight) / 2;

            scrollBar.setVisibleAmount(rows);
            scrollBar.setBlockIncrement(rows - 1);

            for (int i = 0; i < rows; i++) {
                for (int k = 0; k < cols; k++) {
                    int n = ((i * cols) + k);
                    char[] c = { (char) (n + offset) };
                    int startX = k * boxWidth + horizontalBoxPadding +
                        horizontalMarginPadding;
                    //int startY = i * boxHeight + (int) (boxHeight * .7);
                    int startY = i * boxHeight + (boxHeight -
                        verticalBoxPadding - maxDescent +
                        verticalMarginPadding);

                    if (showMetricLines) {
                        // base line
                        g.setColor(baseColor);
                        g2.drawLine(startX - metricsLineExtra,
                                    startY,
                                    startX + metricsLineExtra + maxAdvance,
                                    startY);
                        g2.drawLine(startX,
                                    startY + metricsLineExtra + maxDescent,
                                    startX,
                                    startY - metricsLineExtra - maxAscent);
    
                        // descent line
                        g.setColor(descentColor);
                        g2.drawLine(startX - metricsLineExtra,
                                    startY + maxDescent,
                                    startX + metricsLineExtra + maxAdvance,
                                    startY + maxDescent);
    
                        // ascent line
                        g.setColor(ascentColor);
                        g2.drawLine(startX - metricsLineExtra,
                                    startY - maxAscent,
                                    startX + metricsLineExtra + maxAdvance,
                                    startY - maxAscent);
                    }

                    g.setColor(textColor);
                    g2.drawString(new String(c), startX, startY);
                }
            }
        }

        g.setColor(textColor);
        g2.setFont(numFont);

        int beginY = boxHeight - verticalBoxPadding - maxDescent - verticalMarginPadding;

        // draw the horizontal box lines
        for (int i = 0; i < rows + 1; i++) {
            g.drawLine(
                horizontalMarginPadding,
                i * boxHeight + verticalMarginPadding,
                cols * boxWidth + horizontalMarginPadding,
                i * boxHeight + verticalMarginPadding);
        }

        // draw the vertical box lines
        for (int i = 0; i < cols + 1; i++) {
            g.drawLine(
                i * boxWidth + horizontalMarginPadding,
                verticalMarginPadding,
                i * boxWidth + horizontalMarginPadding,
                rows * boxHeight + verticalMarginPadding);
        }

        // draw the character number
        for (int i = 0; i < rows; i++) {
            for (int k = 0; k < cols; k++) {
                String n = "" + ((i * cols) + k + offset);
                g2.drawString(n, k * boxWidth + horizontalMarginPadding + 2,
                    i * boxHeight + verticalMarginPadding + 12);
            }
        }

    /*
        System.out.print("width: " + getWidth());
        System.out.println("\theight: " + getHeight());
    */
    }

    public Dimension getPreferredSize() {
        /*
        repaint();
        return new Dimension(cols * boxWidth, rowsPerPage * boxHeight);
        */
        return new Dimension(500, 500);
    }

    public Dimension getMinimumSize() {
        return new Dimension(100, 100);
    }

    private Font getNewFont() {
        Font font = null;
        try {
            InputStream fontFileStream = new FileInputStream(fontFileName);
            font = Font.createFont(Font.TRUETYPE_FONT, fontFileStream);
        } catch (FontFormatException e) {
            System.err.println("ERROR: The file \"" + fontFileName + "\" does not appear to be a valid font file");
        } catch (IOException e) {
            System.err.println("ERROR: The file \"" + fontFileName + "\" cannot be found");
        } catch (Exception e) {
            e.printStackTrace();
        }

        Font newfont = null;
        if (font != null) {
            newfont = font.deriveFont(fontSize);
        }

        /*
        FontMetrics metrics = getFontMetrics(font);
        System.out.println("hasUniformLineMetrics(): " + metrics.hasUniformLineMetrics());
        System.out.println("maxDecent(): " + metrics.getMaxDescent());
        System.out.println("maxAscent(): " + metrics.getMaxAscent());
        System.out.println("maxAdvance(): " + metrics.getMaxAdvance());

        maxDescent = metrics.getMaxDescent();
        maxAscent =  metrics.getMaxAscent();
        maxAdvance =  metrics.getMaxAdvance();
        */

        return newfont;
    }

    private static char hexDigits[] =
       {'0', '1', '2', '3',
        '4', '5', '6', '7',
        '8', '9', 'A', 'B',
        'C', 'D', 'E', 'F'
       };

    private String intToUnicode(int n) {
        return "\\u" + hexDigits[(n & 0xF000) >> 12] +
                       hexDigits[(n & 0x0F00) >> 8] +
                       hexDigits[(n & 0x00F0) >> 4] +
                       hexDigits[n & 0x000F];
    }

    public void adjustmentValueChanged(AdjustmentEvent e) {
        offset = e.getValue() * cols;
        repaint();
    }

    public void stateChanged(ChangeEvent e) {
        JSpinner spinner = (JSpinner) e.getSource();
        //System.out.println(spinner.getValue().getClass().getName());

        Integer fontSize = (Integer) spinner.getValue();
        setFontSize(fontSize.floatValue());
        repaint();
    }

    public void itemStateChanged(ItemEvent e) {
        showMetricLines = ! showMetricLines;
        repaint();
    }
}

